import hashlib
import logging
import uuid
from pathlib import Path

import chromadb
from chromadb.config import Settings
from chromadb.utils import embedding_functions
from langchain_text_splitters import RecursiveCharacterTextSplitter
from tqdm import tqdm

log = logging.getLogger(__name__)


def get_content_hash(text):
    """Generate a consistent hash for document content"""
    return hashlib.md5(text.encode("utf-8")).hexdigest()


class Store:
    def __init__(
        self,
        path="/tmp/chroma",
        collection="default",
    ):
        try:
            self.client = chromadb.PersistentClient(
                path=path, settings=Settings(anonymized_telemetry=False)
            )
        except Exception as e:
            log.error(f"Error creating client: {e}")
            return None

        self.embedding_function = embedding_functions.DefaultEmbeddingFunction()

        self.collection = self.client.get_or_create_collection(
            name=collection, embedding_function=self.embedding_function
        )

        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1000,  # number of characters per chunk
            chunk_overlap=200,  # number of characters to overlap between chunks
            length_function=len,
            is_separator_regex=False,
        )

    def add_file(self, filename, parent_id=""):
        if not parent_id:
            parent_id = str(uuid.uuid4())

        try:
            pfile = Path(filename)
            with pfile.open() as f:
                content = f.read()
                parent_hash = get_content_hash(content)
                chunks = self.text_splitter.split_text(content)
                existing = self.collection.get(
                    where={"parent_hash": {"$in": [parent_hash]}},
                )
                if existing["ids"]:
                    log.warning("Duplicate file '%s'", pfile.name)
                    return
                for chunk in tqdm(
                    chunks, desc=f"Processing chunks for {pfile.name}", unit="chunk"
                ):
                    chunk_hash = get_content_hash(chunk)
                    existing = self.collection.get(
                        where={"parent_id,chunk_hash": {"$in": [parent_id, chunk_hash]}},
                    )
                    if existing["ids"]:
                        log.warning("Duplicate chunk on file '%s'", pfile.name)
                        return

                    self.collection.add(
                        documents=[chunk],
                        metadatas=[
                            {
                                "parent_hash": parent_hash,
                                "chunk_hash": chunk_hash,
                                "title": pfile.name,
                                "path": pfile.parent.name,
                                "length": len(content),
                                "parent_id": parent_id,
                            },
                        ],
                        ids=[str(uuid.uuid4())],
                    )
                log.info("Add file '%s'", pfile.name)
        except Exception as e:
            log.error(f"Error adding documents: {e}")

    def search(self, query="", n_results=5):
        result = self.collection.query(
            query_texts=query, n_results=n_results, include=["documents", "metadatas"],
        )
        metadatas = result["metadatas"][0]
        documents = result["documents"][0]
        response = {}
        i = len(metadatas) - 1
        while i >= 0: # group by parent_id
            parent_id = metadatas[i]["parent_id"]
            response[parent_id] = {
                "content": documents[i],
                "meta": metadatas[i],
                "parent_id": parent_id,
                "order": i,
            }
            i -= 1
        response_arr = [(r["order"], r) for r in response.values()] # sort by order
        response_arr = sorted(response_arr, key=lambda x: x[0])
        return [item for _, item in response_arr]

